from ksl_util.log import printlog

def translate(string):
    
    batch_list = [
        'batch_size', 'batch', 'batchSize', 'BatchSize', 'batch size', 'batchsize'
    ]
    
    img_size_list = [
        'img_size', 'image_size', 'imgSize', 'ImageSize', 'Image_Size', 'image size', 'imagesize', 'imgsize'
    ]
    
    img_resize_flag_list = [
        'imageresizeflag', 'imgresizeflag', 'imgflag', 'imgresizingflag', 'resizeflag'
    ]
    
    train_flag_list = [
        'train', 'trn', 'training'
    ]
    
    valid_flag_list = [
        'valid', 'vld', 'validation'
    ]
    
    test_flag_list = [
        'test', 'tst', 'testing'
    ]
    
    
    batch = 'batch_size'
    img_size = 'image_size'
    img_resize_flag = 'image_resize_flag'
    train_flag = 'trn'
    valid_flag = 'vld'
    test_flag = 'tst'


    string_tmp = string.lower()
    string_tmp = string_tmp.replace('_', '')
    
    
    if string_tmp in batch_list: return batch
    elif string_tmp in img_size_list: return img_size
    elif string_tmp in img_resize_flag_list: return img_resize_flag
    elif string_tmp in train_flag_list: return train_flag
    elif string_tmp in valid_flag_list: return valid_flag
    elif string_tmp in test_flag_list: return test_flag
    else: return string
    


class translated_dictionary:
    
    def __init__(self, init_value=None):
        self.__dictionary__ = dict()
        self.__keys__ = list()
        
        if init_value is not None:
            self.translate(init_value)
        
    
    def translate(self, dictionary):
        
        for key in dictionary.keys():
            self.__keys__.append(translate(key))
            self.__dictionary__[translate(key)] = dictionary[key]
            
    
    def __getitem__(self, item):
        return self.__dictionary__[translate(item)]
    
    def keys(self):
        return self.__keys__
    


from ksl_util.time.time_check import current_time
import inspect, re

def printlog(*args, **kwargs):
    time = current_time()
    end = '\n'
    method_name = ''

    if kwargs.__contains__('end'):
        end = kwargs['end']

    if kwargs.__contains__('method') and kwargs['method'] == True:
        print('{:<30} :'.format(time), *args)
        print('{:<30} :'.format(inspect.stack()[1][2]), inspect.stack()[1][3], end=end)
        # print(inspect.stack())
        return

    if 'save_txt' in kwargs:
        file = kwargs['save_txt']
        txt = '{:<30} :'.format(time)
        for arg in args:
            txt += arg
        txt += end
        file.write('%s' % txt)

    print('{:<30} :'.format(time), *args, end=end)


def printcommand(command, *args, **kwargs):
    
    printlog('{:<15}'.format('(%s)' % command), *args)


def class_check(_object, _class):

    assert _class.__name__ == _object.__class__.__name__ or _class.__name__ in [name.__name__ for name in _object.__class__.__bases__], 'object of <%s> is not a class of <%s>' % (_object.__class__.__name__, _class.__name__)


def print_varname(var_name):
    print('{:<30}'.format(var_name), eval(var_name))




import multiprocessing
import os

import cv2
import numpy as np

from ksl_util.log import printlog
from ksl_util.multiprocessing.mp_class import SingleProcess, Processes
from ksl_util.time.time_check import TimeCheck

import ctypes

import platform


class Loader:

    def __init__(self):
        self.path = None
        self.names = None
        self.index = 0
        self.index_random = 0
        self.random_indexes = None
        self.current_images = list()


        ###### for testing
        self.flagg = False


    def load_path(self, path):

        assert os.path.exists(path)

        self.path = path
        self.names = os.listdir(path)
        self.random_indexes = np.asarray(range(0, len(self.names)))

        np.random.shuffle(self.random_indexes)

        return self


    def load_images(self, numImages=None, grayscale=False, printable=True, random=False, callbacks=None):

        if self.flagg:
            return self.current_images

        self.current_images.clear()
        if numImages is None:
            numImages = self.__len__()

        while len(self.current_images) != numImages:
            if random:
                image = cv2.imread('%s/%s' % (self.path, self.names[self.random_indexes[self.index_random]]))
            else:
                image = cv2.imread('%s/%s' % (self.path, self.names[self.index]))

            if grayscale:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
                image = np.expand_dims(image, axis=3)
            if printable: print('%s' % self.names[self.random_indexes[self.index_random]], end=', ')

            self.current_images.append(image)

            if random:
                self.index_random = (self.index_random + 1) % self.__len__()
                if self.index_random == 0:
                    np.random.shuffle(self.random_indexes)
            else:
                self.index = (self.index + 1) % self.__len__()

        if printable: print('')

        if callbacks is not None:
            self.current_images = callbacks(self.current_images)

        self.flagg = True
        return self.current_images



    def __len__(self):
        return len(self.names)


    def __getitem__(self, item):
        return self.current_images[item]



class PairLoader:

    def __init__(self):
        self.path1                                      = None
        self.path2                                      = None
        self.names1                                     = None
        self.names2                                     = None
        self.random_indexes                             = None
        self.current_images                             = list()

        self.index                                      = 0
        self.index_random                               = 0


    def load_path(self, path1, path2):

        assert os.path.exists(path1)
        assert os.path.exists(path2)

        self.path1                                      = path1
        self.path2                                      = path2

        self.names1                                     = os.listdir(path1)
        self.names2                                     = os.listdir(path2)
        
        self.names1                                     .sort()
        self.names2                                     .sort()

        self.random_indexes                             = np.asarray(range(0, len(self.names1)))

        assert len(self.names1) == len(self.names2)
        np.random.shuffle(self.random_indexes)


        return self


    def load_images(self, numImages=None, grayscales=(False, True), printable=True, random=False, callbacks=None):

        assert len(grayscales) == 2

        self.current_images.clear()

        if numImages is None:
            numImages = len(self.names1)

        while len(self.current_images) != numImages:
            if random:
                image1 = cv2.imread('%s/%s' % (self.path1, self.names1[self.random_indexes[self.index_random]]))
                image2 = cv2.imread('%s/%s' % (self.path2, self.names2[self.random_indexes[self.index_random]]))
            else:
                image1 = cv2.imread('%s/%s' % (self.path1, self.names1[self.index]))
                image2 = cv2.imread('%s/%s' % (self.path2, self.names2[self.index]))


            if grayscales[0]:
                image1 = cv2.cvtColor(image1, cv2.COLOR_BGR2GRAY)
                image1 = np.expand_dims(image1, axis=3)
            if grayscales[1]:
                image2 = cv2.cvtColor(image2, cv2.COLOR_BGR2GRAY)
                image2 = np.expand_dims(image2, axis=3)


            if printable:
                if random:
                    print('%s, %s' % (self.names1[self.random_indexes[self.index_random]], self.names2[self.index]), end='| ')
                else:
                    print('%s, %s' % (self.names1[self.index], self.names2[self.index]), end='| ')

            self.current_images.append((image1, image2))


            if random:
                self.index_random = (self.index_random + 1) % self.__len__()
                if self.index_random == 0:
                    np.random.shuffle(self.random_indexes)
            else:
                self.index = (self.index + 1) % self.__len__()

        if printable: print('')

        if callbacks is not None:
            self.current_images = callbacks(self.current_images)

        return self.current_images


    def __len__(self):
        assert len(self.names2) == len(self.names1)
        return len(self.names1)


    def __getitem__(self, item):
        return self.current_images[item]


class __ImageHolder__(SingleProcess):

    def __init__(self):
        super().__init__()
        self.loader                                     = None
        self.holder                                     = multiprocessing.Queue()

        self.buffer_size                                = 100
        self.capacity                                   = 5
        self.grayscale                                  = False
        self.random                                     = False


    def setBufferSize(self, size):
        self.buffer_size                                = size
        return self

    def setCapacity(self, capacity):
        self.capacity                                   = capacity
        return self

    def setGrayScale(self, grayscale):
        self.grayscale                                  = grayscale
        return self

    def setPath(self, *path):
        pass

    def setRandom(self, random):
        self.random                                     = random
        return self


    def __job__(self, queue):
        while queue.qsize() < self.buffer_size:
            images = self.__load__images__()
            for image in images:
                queue.put(image)


    def __routine__(self, queue):
        while True:
            if queue.qsize() >= self.buffer_size: continue
            self.__job__(queue)


    def __init__process__(self):
        p = multiprocessing.Process(target=self.__routine__, args=[self.holder])
        p.daemon = True
        return p

    def load(self):
        Processes()\
        .append(self.build())\
        .start()

        while self.holder.qsize() < self.capacity:
            pass

    def get(self):
        item = self.holder.get()
        return item

    def get_items(self, numImages=1):

        while self.holder.qsize() <= self.capacity:
            pass
        tmp = list()
        image = self.get()
        while len(tmp) != numImages:
            tmp.append(image)
            del image
            image = self.get()
        del image

        return tmp

    def clear(self):
        self.holder.close()

    def __load__images__(self):
        pass

    def __len__(self):
        return self.holder.qsize()


class ImageHolder(__ImageHolder__):

    def __init__(self):
        super().__init__()
        self.loader = Loader()

    def setPath(self, *path):
        assert len(path) == 1
        self.loader.load_path(path[0])
        return self

    def __load__images__(self):
        return self.loader.load_images(numImages=self.capacity, grayscale=self.grayscale, printable=False, random=self.random)


class ImagePairHolder(__ImageHolder__):

    def __init__(self):
        super().__init__()
        self.loader = PairLoader()
        self.grayscale = (False, True)

    def setPath(self, *path):
        assert len(path) == 2
        assert os.path.exists(path[0]) and os.path.exists(path[1]), '%s %s' % (path[0], path[1])
        self.loader.load_path(path[0], path[1])
        return self

    def setGrayScale(self, grayscale):
        assert len(grayscale) == 2
        self.grayscale = grayscale
        return self

    @TimeCheck
    def __load__images__(self):
        return self.loader.load_images(numImages=self.capacity, grayscales=self.grayscale, printable=False, random=self.random)




def imread3(paths, flags=None):
    
    assert flags is None or len(paths) == len(flags) or len(flags) == 1
    
    tmp = list()
    
    for i in range(0, len(paths)):
        if flags is None:
            tmp.append(cv2.imread(paths[i], cv2.IMREAD_COLOR))
        elif len(flags) == 1:
            tmp.append(cv2.imread(paths[i], flags[0]))
        else:
            tmp.append(cv2.imread(paths[i], flags[i]))
    
    return tuple(tmp)


if platform.system() == 'Darwin':
    from AppKit import NSScreen

elif platform.system() != 'Windows':
    from Xlib.display import Display
    MAX_X = Display(':0').screen().width_in_pixels


def imshow2(images, nameflag='img', windowFlag=None, x=0, y=0, margin=10):

    if windowFlag is not None:
        assert len(images) == len(windowFlag)

    currentX = x
    currentY = y

    if platform.system() == 'Windows':
        maxX = ctypes.windll.user32.GetSystemMetrics(0)
    elif platform.system() == 'Darwin':
        maxX = NSScreen.mainScreen().frame().size.width
        margin += 120
    else:
        maxX = MAX_X
        margin += 120

    for index, image in enumerate(images):
        if windowFlag is None:
            cv2.imshow('%s%d' % (nameflag, index), image)
            cv2.moveWindow('%s%d' % (nameflag, index), currentX, currentY)
        else:
            cv2.imshow('%s_%s%d' % (windowFlag[index], nameflag, index), image)
            cv2.moveWindow('%s_%s%d' % (windowFlag[index], nameflag, index), currentX, currentY)
            

        currentX += margin + image.shape[1]
        if currentX + image.shape[1] > maxX:
            currentX = 0
            currentY += margin + image.shape[0] + 25



def image_modification(images, function, *args, **kwargs):

    tmp = list()
    
    for image in images:
        tmp.append(np.asarray(function(image, *args, **kwargs), dtype=np.uint8))
        
    return tuple(tmp)



from skimage import io
import cv2
import numpy as np

def read_tif(file_name, alpha=0.3):

    images = io.imread(file_name)
    tmp = []
    for image in images:
        tmp.append(cv2.convertScaleAbs(image, alpha=alpha))
    return np.asarray(tmp)




from sklearn import metrics

def auc(X, Y):
    return 1 / (len(X) * len(Y)) * sum([kernel(x, y) for x in X for y in Y])


def kernel(X, Y):
    return .5 if Y == X else int(Y < X)


def structural_components(X, Y):
    V10 = [1 / len(Y) * sum([kernel(x, y) for y in Y]) for x in X]
    V01 = [1 / len(X) * sum([kernel(x, y) for x in X]) for y in Y]
    return V10, V01


def get_S_entry(V_A, V_B, auc_A, auc_B):
    return 1 / (len(V_A) - 1) * sum([(a - auc_A) * (b - auc_B) for a, b in zip(V_A, V_B)])


def z_score(var_A, var_B, covar_AB, auc_A, auc_B):
    return (auc_A - auc_B) / ((var_A + var_B - 2 * covar_AB) ** (.5))


def group_preds_by_label(preds, actual):
    X = [p for (p, a) in zip(preds, actual) if a]
    Y = [p for (p, a) in zip(preds, actual) if not a]
    return X, Y


def cal_auc(pred, label):
    X_A, Y_A = group_preds_by_label(pred, label)
    auc_A = auc(X_A, Y_A)
    return auc_A


def draw_roc(pred, label, name='ROC'):
    fpr, tpr, thresholds = metrics.roc_curve(label, pred)
    roc_auc = metrics.auc(fpr, tpr)
    display = metrics.RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=roc_auc, estimator_name=name)
    return display



import numpy as np


def random_crop(image, new_shape=None, fh=1., fw=1., onlyIndex=True):

    # assert len(image.shape) == 3
    assert new_shape is None or len(new_shape) == 2

    image_width, image_height = image.shape[:2]

    if (fw != 1. or fh != 1.) and new_shape is None:
        new_shape = (int(image_width * fh), int(image_height * fw))

    if new_shape is None:
        new_shape = (image_width, image_height)


    new_width, new_height = new_shape
    sub_width, sub_height = image_width - new_width, image_height - new_height
    width_range, height_range = np.asarray(range(0, sub_width + 1)), np.asarray(range(0, sub_height + 1))

    width, height = np.random.choice(width_range, size=1, replace=False)[0], np.random.choice(height_range, size=1, replace=False)[0]

    if onlyIndex:
        return CropObject((width, width + new_width, height, height + new_height))
    else:
        if len(image.shape) == 3:
            return image[width:width+new_width, height:new_height, :]
        else:
            return image[width:width+new_width, height:new_height + height]




class CropObject:

    def __init__(self, point):
        self.point = point

    def crop(self, image):
        assert len(image.shape) == 3 or len(image.shape) == 2

        point = self.point
        
        if len(image.shape) == 3:
            return image[point[0]:point[1], point[2]:point[3], :]
        else:
            return image[point[0]:point[1], point[2]:point[3]]

    def __repr__(self):
        return '(%d, %d, %d, %d)' % (self.point[0], self.point[1], self.point[2], self.point[3])




import cv2
import os
import math
import numpy as np

from ksl_util.file.image.image_loader import ImagePairHolder, PairLoader
from ksl_util.log import printlog


def miou(pred, anno):

    tp = np.logical_and(pred, anno)
    tp = np.asarray(tp, 'float64')
    tp = np.sum(tp)

    fp = np.logical_and(np.logical_not(anno), pred)
    fp = np.asarray(fp, 'float64')
    fp = np.sum(fp)

    fn = np.logical_and(np.logical_not(pred), anno)
    fn = np.asarray(fn, 'float64')
    fn = np.sum(fn)

    tn = np.logical_and(np.logical_not(pred), np.logical_not(anno))
    tn = np.asarray(tn, 'float64')
    tn = np.sum(tn)

    return tp, fp, fn, tn


def sensitivity(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TP / (TP + FN) if TP + FN != 0 else None


def specificity(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TN / (TN + FP) if TN + FP != 0 else None


def precision(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TP / (TP + FP) if TP + FP != 0 else None


def recall(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TP / (TP + FN) if TP + FN != 0 else None


def accuracy(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return (TP + TN) / (TP + TN + FP + FN) if TP + FN + FP + TN != 0 else 1.0


def ppv(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TP / (TP + FP) if TP + FN != 0 else None


def npv(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return TN / (TN + FN) if TN + FN != 0 else None


def f1_score(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return (TP + TP) / (TP + TP + FP + FN) if TP + FN + FP + TN != 0 else 1.0


def balanced_accuracy(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return (sensitivity(TP, FP, FN, TN) + specificity(TP, FP, FN, TN)) / 2.0


def MCC(TP, FP, FN, TN):
    TP, FP, FN, TN = tuple([float(p) for p in [TP, FP, FN, TN]])
    return (TP * TN - FP * FN) / math.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) if (TP + FP) * (TP + FN) * (
            TN + FP) * (TN + FN) != 0 else None


def all_evaluations(TP, FP, FN, TN):
    results = dict()
    for func in [
        sensitivity, specificity, precision, recall, ppv, npv, f1_score, accuracy, balanced_accuracy, MCC
    ]:
        results[func.__name__] = func(TP, FP, FN, TN)
    return results



def save_evaluate(path_image, path_ground, absolute_file_name):

    num_images, num_ground = len(os.listdir(path_image)), len(os.listdir(path_ground))

    assert num_images == num_ground, 'Different number of images'

    loader = PairLoader()
    loader.load_path(path_image, path_ground)

    images = loader.load_images(num_images, grayscales=(True, True), printable=True, random=False)
    printlog('loaded!!')

    tTP = 0
    tFP = 0
    tFN = 0
    tTN = 0
    tmatched = 0
    index = 0

    with open(absolute_file_name, 'w') as save_file_name:
        printlog(save_file_name)
        save_file_name.write('index, iou, acc, TP, FP, FN, TN')
        save_file_name.write('\n')

        for image in images:
            img, gnd = image[0], image[1]
            matched, TP, FP, FN, TN = miou(img, gnd)

            TP += 1e-15
            tTP += TP
            tFP += FP
            tFN += FN
            tTN += TN
            tmatched += matched

            message = '%d, %05f, %05f, %d, %d, %d, %d\n' % (index, float(TP) / (TP + FP + FN), float(matched) / (TP + TN + FN + FP), TP, FP, FN, TN)
            save_file_name.write(message)
            index += 1

        message = '%s, %05f, %05f, %d, %d, %d, %d\n' % ('mean', float(tTP) / (tTP + tFP + tFN), float(tmatched) / (tTP + tTN + tFN + tFP), tTP, tFP, tFN, tTN)
        printlog(message)
        save_file_name.write(message)

    print('fin.\n\n')



import os
import shutil
from multiprocessing import Process, Queue
from pathlib import Path

import cv2
import numpy as np
from ksl_util.file.directory.directory_manager import mkdir
from ksl_util.log import printlog
from ksl_util.time.time_check import TimeCheck


class ImageReader:

    def __init__(self, directory, number=1, grayscale=False, reverse=False):

        self.directory                      = directory
        self.index                          = -1
        self.grayscale                      = grayscale
        self.nameList                       = os.listdir(directory)
        self.indexes                        = [i for i in range(0, number)]
        self.max_length                     = number
        self.reverse                        = reverse

        self.nameList.sort()
        if reverse:
            self.nameList.reverse()
        np.random.shuffle(self.indexes)

    def setOrder(self, numberList):

        assert len(numberList) == len(self.indexes)

        self.indexes = numberList


    def changeDirectory(self, directory):
        self.directory                      = directory
        self.nameList                       = os.listdir(directory)
        self.nameList.sort()
        if self.reverse: self.nameList.reverse()



    def read(self, callback=None, withName=False):

        self.index = (self.index + 1) % self.max_length
        # printlog("::::: %d" % self.max_length)
        index = self.indexes[self.index] % len(self.nameList)

        image = None
        while image is None or len(image.shape) != 3:
            image = cv2.imread('%s/%s' % (self.directory, self.nameList[index]), cv2.IMREAD_COLOR)
            # print('%s/%s' % (self.directory, self.nameList[index]))

        if self.grayscale:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        if callback is not None:
            image = callback(image)


        if withName:
            return image, '%s/%s' % (self.directory, self.nameList[index])
        else:
            return image


class ImagePairReader:

    def __init__(self, directory1, directory2, reverse=False):

        assert len(directory1) == 3 and len(directory2) == 3
        assert os.path.exists(directory1[0]), directory1[0]
        assert os.path.exists(directory2[0]), directory2[0]
        assert isinstance(directory1[1], int) and isinstance(directory2[1], int)
        assert isinstance(directory1[2], bool) and isinstance(directory2[2], bool)
        assert directory1[1] == directory2[1]

        self.iR1 = ImageReader(directory=directory1[0], number=directory1[1], grayscale=directory1[2], reverse=reverse)
        self.iR2 = ImageReader(directory=directory2[0], number=directory2[1], grayscale=directory2[2], reverse=reverse)

        self.indexes = [i for i in range(0, directory1[1])]
        np.random.shuffle(self.indexes)

        self.iR1.setOrder(self.indexes)
        self.iR2.setOrder(self.indexes)


    def changeDirectory(self, directory1, directory2):
        self.iR1.changeDirectory(directory1)
        self.iR2.changeDirectory(directory2)


    def changeOrder(self):
        np.random.shuffle(self.indexes)

        self.iR1.setOrder(self.indexes)
        self.iR2.setOrder(self.indexes)



    def read(self, callback=None, withName=False):

        img1, name1 = self.iR1.read(callback=callback, withName=True)
        img2, name2 = self.iR2.read(callback=callback, withName=True)
        # printlog(name1, name2)

        tmp1, tmp2 = name1, name2
        name1 = name1.split('/')
        name1 = name1[len(name1)-1].replace('.jpeg', '.png').replace('.jpg', '.png')
        name2 = name2.split('/')
        name2 = name2[len(name2)-1]
        assert name1 == name2, '%s :: %s' % (tmp1, tmp2)

        if withName:
            return (img1, name1), (img2, name2)
        else:
            return img1, img2


class ImageCopier:


    def __init__(self, src, dst, number, callback=None, arg=None, reverse=False):

        assert os.path.exists(src)
        assert os.path.exists(dst)
        assert isinstance(number, int) or isinstance(number, np.int32)

        self.src                            = src
        self.dst                            = dst
        self.number                         = number

        self.callback                       = callback
        self.arg                            = arg
        self.reverse                        = reverse

    def __copyProcess__(self):

        # get file names
        file_names = os.listdir(self.src)
        file_names.sort()
        if self.reverse: file_names.reverse()
        file_names = file_names[0:min(self.number, len(file_names))]

        # copies
        for file_name in file_names:
            file_name = file_name.replace('jpg', 'png')
            shutil.copy('%s/%s' % (self.src, file_name), '%s/%s' % (self.dst, file_name))

        if self.callback is not None:
            self.callback(self.arg)


    def build(self):
        self.process = Process(target=self.__copyProcess__)
        self.process.daemon = True
        self.process.start()


    def join(self):
        self.process.join()



class ImageQueue:

    def __init__(self, dir1, dir2, totalNum, capacity, batch=1, grayscale=(False, True), callback=None, reverse=False):

        assert len(grayscale) == 2

        self.queue                              = Queue()
        self.pathQueue                          = Queue()

        self.dir1                               = (dir1, totalNum, grayscale[0])
        self.dir2                               = (dir2, totalNum, grayscale[1])
        self.totalNum                           = totalNum

        self.capacity                           = capacity
        self.batch                              = batch

        self.callback                           = callback

        self.process                            = Process(target=self.__Process__, args=[self.queue, self.pathQueue, self.dir1, self.dir2, callback, reverse])
        self.process.daemon                     = True
        self.process                            .start()

    def changeDirectories(self, dir1, dir2):
        self.pathQueue.put((dir1, dir2))


    def stop(self):
        self.process.kill()


    def get(self, callback=None):

        imgs, gnds = list(), list()

        while len(imgs) != self.batch:
            img, gnd = self.queue.get()
            if callback is not None:
                img, gnd = callback(img, gnd)

            imgs.append(img)
            gnds.append(gnd)

        return np.asarray(imgs), np.asarray(gnds)


    def currentSize(self):
        return self.queue.qsize()


    def __Process__(self, queue, pathQueue, info1, info2, callback, reverse):

        IPL = ImagePairReader(directory1=info1, directory2=info2, reverse=reverse)
        number = 0

        while True:

            if pathQueue.qsize() > 0:
                path1, path2 = pathQueue.get()
                IPL.changeDirectory(path1, path2)

            if queue.qsize() > self.capacity: continue

            queue.put(IPL.read(callback=callback))
            number += 1

            if number >= self.totalNum:
                number = 0
                IPL.changeOrder()

    def __len__(self):
        return self.queue.qsize()




class ImageBatchLoader:

    def __init__(self, dir1, dir2, totalNum, capacity, batch=1, grayscale=(False, True), callback=None, save_flag=True, save_dir='.', reverse=False):

        # make tmp directory
        self.tmpPath                        = '%s/Desktop/.tf_tmp' % Path.home()
        self.tmpPath                        = mkdir(self.tmpPath)
        self.tmpPath                        = '%s/%s' % (self.tmpPath, save_dir)

        if os.path.exists(self.tmpPath): shutil.rmtree(self.tmpPath, ignore_errors=True)

        self.queue                          = Queue()
        self.imgPath                        = mkdir('%s/img' % self.tmpPath)
        self.gndPath                        = mkdir('%s/gnd' % self.tmpPath)

        self.imageQueue                     = ImageQueue(dir1, dir2, totalNum, capacity, batch, grayscale, callback, reverse)
        self.imgCopier                      = ImageCopier(dir1, self.imgPath, totalNum, callback=self.__copyCallback__, arg=self.queue, reverse=reverse)
        self.gndCopier                      = ImageCopier(dir2, self.gndPath, totalNum, callback=self.__copyCallback__, arg=self.queue, reverse=reverse)

        if save_flag:
            self.imgCopier.build()
            self.gndCopier.build()

            thread                              = Process(target=self.__checkThread__, args=(self.queue,))
            thread.daemon                       = True
            thread                              .start()

    def deleteDir(self):
        shutil.rmtree(self.tmpPath, ignore_errors=True)


    def get(self, callback=None):
        return self.imageQueue.get(callback=callback)

    # run in another process
    def __copyCallback__(self, queue):
        queue.put('1')


    def __checkThread__(self, queue):

        while queue.qsize() < 2:
            continue

        self.imageQueue.changeDirectories(self.imgPath, self.gndPath)


import numpy as np
import cv2

def cal_IoU(img, gnd):
    img = img > 0
    gnd = gnd > 0
    
    tp = np.logical_and(img, gnd)
    fp = np.logical_and(np.logical_not(gnd), img)
    fn = np.logical_and(gnd, np.logical_not(img))
    tn = np.logical_and(np.logical_not(img), np.logical_not(gnd))
    
    tp = tp.astype(np.float32)
    fp = fp.astype(np.float32)
    fn = fn.astype(np.float32)
    tn = tn.astype(np.float32)
    
    tp = np.sum(tp)
    fp = np.sum(fp)
    fn = np.sum(fn)
    tn = np.sum(tn)
    
    iou = tp / (tp + fp + fn) if tp + fp + fn != 0 else 1.0
    
    return iou, tp, fp, fn, tn



if __name__ == '__main__':


    path_img = 'N:\\SPIE\\prediction_paper\\Fluo-N2DL-HeLa\\FusionNet_batch=1_size=384'
    path_gnd = 'R:\\Resource\\Fluo-N2DL-HeLa\\training_images2\\gnd'
    
    img = cv2.imread('%s/0000.png' % path_img)
    gnd = cv2.imread('%s/01_t013.png' % path_gnd)

    print(cal_IoU(img, gnd))



from ksl_util import time
from functools import wraps


def current_time():
    now = time.localtime()
    r_time = "%04d%02d%02d__%02d%02d%02d" % (now.tm_year, now.tm_mon, now.tm_mday, now.tm_hour, now.tm_min, now.tm_sec)

    return r_time

def current_date():
    now = time.localtime()
    r_time = "%04d%02d%02d" % (now.tm_year, now.tm_mon, now.tm_mday)

    return r_time

class time_checker:
    def __init__(self):
        self.__start__      = 0
        self.__end__        = 0

    def start(self):
        self.__start__ = time.time()

    def end(self, printable=True):
        self.__end__ = time.time()
        timer = self.__end__ - self.__start__
        if printable: print('time : ', timer, '(s)')
        return timer


def TimeCheck(func):
    def wrapeer(func):
        @wraps(func)
        def decorated(*args, **kwargs):
            tc = time_checker()
            tc.start()
            result = func(*args, **kwargs)
            print('{:<30} : '.format(func.__name__), tc.end(printable=False), '(s)')
            return result
        return decorated
    return wrapeer(func)

def Sleep(__time):
    def wrapeer(func):
        @wraps(func)
        def decorated(*args, **kwargs):
            time.sleep(__time)
            result = func(*args, **kwargs)
            return result
        return decorated
    return wrapeer





import os
import numpy as np
from os.path import expanduser
from ksl_util.time.time_check import current_date


def mkdir(dir_name, makeDir=True):
    if not os.path.exists(dir_name) and makeDir:
        os.makedirs('%s' % (dir_name))

    return os.path.abspath(dir_name)


def getFileList(directory, absolute=False, callback=None):

    tmp = os.listdir(directory)
    tmp.sort()
    if not absolute:
        return tmp

    lists = []
    for name in tmp:
        abName = '%s/%s' % (directory, name)
        if callback is not None:
            lists.append((abName, callback(abName)))
        else:
            lists.append(abName)

    return lists


def getFileLists(path1, path2, sort=True, absolute=False):
    
    tmp1 = os.listdir(path1)
    tmp2 = os.listdir(path2)
    
    if sort:
        tmp1.sort()
        tmp2.sort()
    
    if absolute:
        ab1 = ['%s/%s' % (path1, name) for name in tmp1]
        ab2 = ['%s/%s' % (path2, name) for name in tmp2]
        
        tmp1, tmp2 = ab1, ab2
    
    return tmp1, tmp2

def getFileLists3(paths, sort=True, absolute=False, number=None):

    tmp = list()
    
    for path in paths:
        tmp1 = os.listdir(path)
        tmp2 = []
        for t in tmp1:
            if t.find('.') == 0: continue
            tmp2.append(t)
        tmp1 = tmp2
        if sort:
            tmp1.sort()
    
        if absolute:
            ab1 = [os.path.abspath('%s/%s' % (path, name)) for name in tmp1]
            tmp1 = ab1

        if number is not None:
            tmp1 = tmp1[0:number]

        tmp.append(tmp1)

    return tuple(tmp)

from tqdm import tqdm
def checkNPZ(path, names):
    
    flag = True
    return_values = []
    return_values.append(False)
    for name in names:
        flag &= os.path.exists('%s/%s00.npz' % (path, name))
        print(flag, path, name)
        return_values.append(None)

    if not flag:
        return tuple(return_values)
    
    print('here')

    return_values.clear()
    return_values.append(True)
    for name in tqdm(names, desc='npz reading'):
        index = 0
        tmps = []
        error_count = 0
        while os.path.exists('%s/%s%02d.npz' % (path, name, index)):
            print('%s/%s%02d.npz' % (path, name, index))
            try:
                tmps.append(np.load('%s/%s%02d.npz' % (path, name, index))[name])
                index += 1
            except Exception as ex:
                error_count += 1
                print('error in %s_%d' % (name, index))
                
                if error_count >= 10:
                    return_values.clear()
                    return_values.append(False)
                    for name in names:
                        flag &= os.path.exists('%s/%s00.npz' % (path, name))
                        return_values.append(None)
                    return tuple(return_values)
        tmps = np.concatenate(tmps, axis=0)
        return_values.append(tmps)

    return_values = tuple(return_values)
    return tuple(return_values)


def search_files(root, ext: list = ['jpg', 'png'], useCache=True, refresh=False):
    r_values = []

    tmp_directory = mkdir('%s/.tmp_ksl' % expanduser('~'), makeDir=useCache)
    tmp_directory = mkdir('%s/search_files' % tmp_directory, makeDir=useCache)
    tmp_directory = mkdir('%s/%s__%s' % (tmp_directory, current_date(), os.getcwd().replace(':', '').replace(":\/\/", "__").replace(':\\\\', '__').replace('\/', '__').replace('\\', '__')), makeDir=useCache)
    file_name = os.path.basename(__file__).replace('py', 'txt')
    file_name = '%s/%s' % (tmp_directory, file_name)

    if refresh and os.path.exists(file_name):
        os.remove(file_name)

    if useCache and os.path.exists(file_name):
        line = []
        with open(file_name, 'r') as f:
            txt = f.readline()
            while txt:
                txt = txt.rstrip().lstrip()
                line.append(txt)
                txt = f.readline()
        return line

    for (root, dirs, files) in os.walk(root):
        r_values.extend(['%s/%s' % (root, file) for file in files if os.path.splitext(file)[1].replace('.', '') in ext])

    r_values = [value for value in r_values if os.path.splitext(value)[1] in ['.%s' % e for e in ext]]

    if useCache:
        with open(file_name, 'w') as f:
            for r_value in r_values:
                f.write('%s\n' % os.path.abspath(r_value))

    return r_values


